import sqlite3
import zlib
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES
from spectrum_fundamentals.mod_string import internal_to_mod_mass, internal_without_mods

from .spectral_library import SpectralLibrary

DLIB_COL_NAMES = [
    "MassArray",
    "IntensityArray",
    "MassEncodedLength",
    "IntensityEncodedLength",
    "PrecursorCharge",
    "PeptideModSeq",
    "PeptideSeq",
    "RTInSeconds",
    "PrecursorMz",
]


class DLib(SpectralLibrary):
    """Main to init a DLib obj."""

    def _calculate_masked_values(self, fragmentmz: List[np.ndarray], intensities: List[np.ndarray]):
        """
        Internal function that masks, filters, byte encodes, swaps and compresses fragmentmz \
        and intensities.

        This will produce the data for the following columns in this order:
            - 'MassArray'
            - 'IntensityArray',
            - 'MassEncodedLength',
            - 'IntensityEncodedLength'.
        :param fragmentmz: fragmentmz provided in __init__
        :param intensities: intensities provided in __init__
        :return: 4 lists as described above
        """
        mz_bytes_list = []
        i_bytes_list = []
        mz_lengths = []
        i_lengths = []
        for mz, i in zip(fragmentmz, intensities):
            # mask to only existing peaks
            mask = i >= self.min_intensity_threshold
            sort_index = np.argsort(mz[mask])
            masked_mz_ordered = mz[mask][sort_index]
            masked_i_ordered = i[mask][sort_index]

            # convert arrays to big-endian byte blops
            masked_mz_ordered = np.array(masked_mz_ordered, dtype=np.double)
            masked_i_ordered = np.array(masked_i_ordered * 100, dtype=np.float32)
            masked_mz_ordered.byteswap(inplace=True)
            masked_i_ordered.byteswap(inplace=True)
            bytes_mz = bytes(masked_mz_ordered)
            bytes_i = bytes(masked_i_ordered)
            mz_bytes_list.append(zlib.compress(bytes_mz))
            i_bytes_list.append(zlib.compress(bytes(bytes_i)))
            mz_lengths.append(len(bytes_mz))
            i_lengths.append(len(bytes_i))
        return mz_bytes_list, i_bytes_list, mz_lengths, i_lengths

    @staticmethod
    def create_database(path: Union[str, Path]):
        """
        Creates the database file with prefab tables entries, peptidetoprotein (p2p) and metadata, according to the \
        dlib specification.

        :param path: specifies the path of the created database file
        """
        sql_create_entries = """
            CREATE TABLE entries
            (   PrecursorMz double not null,
                PrecursorCharge int not null,
                PeptideModSeq string not null,
                PeptideSeq string not null,
                Copies int not null,
                RTInSeconds double not null,
                Score double not null,
                MassEncodedLength int not null,
                MassArray blob not null,
                IntensityEncodedLength int not null,
                IntensityArray blob not null,
                CorrelationEncodedLength int,
                CorrelationArray blob,
                RTInSecondsStart double,
                RTInSecondsStop double,
                MedianChromatogramEncodedLength int,
                MedianChromatogramArray blob,
                SourceFile string not null
            )
        """
        sql_create_p2p = """
            CREATE TABLE peptidetoprotein
            (PeptideSeq string not null, isDecoy boolean, ProteinAccession string not null)
        """
        sql_create_meta = """
            CREATE TABLE metadata (Key string not null, Value string not null)
        """
        sql_insert_meta = "INSERT INTO metadata VALUES (?,?)"
        conn = sqlite3.connect(path)
        c = conn.cursor()
        c.execute(sql_create_entries)
        c.execute(sql_create_p2p)
        c.execute(sql_create_meta)
        c.execute(sql_insert_meta, ["version", "0.1.14"])
        c.execute(sql_insert_meta, ["staleProteinMapping", "true"])
        conn.commit()

    def write(self):
        """Writes the entries ad p2p table to file."""
        self._write_entries(index=False, if_exists="append", method="multi", chunksize=self.chunksize)
        self._write_p2p(index=False, if_exists="append", method="multi", chunksize=self.chunksize)

    def _write_entries(self, *args, **kwargs):
        """
        Internal function to write the entries table.

        :param args: forwarded to pandas.to_sql
        :param kwargs: forwarded to pandas.to_sql
        """
        conn = sqlite3.connect(self.out_path)
        self.entries.to_sql(name="entries", con=conn, *args, **kwargs)
        conn.commit()

    def _write_p2p(self, *args, **kwargs):
        """
        Internal function to write the p2p table.

        :param args: forwarded to pandas.to_sql
        :param kwargs: forwarded to pandas.to_sql
        """
        conn = sqlite3.connect(self.out_path)
        self.p2p.to_sql(name="peptidetoprotein", con=conn, *args, **kwargs)
        conn.commit()

    def prepare_spectrum(self):
        """Converts grpc output and metadata dataframe into dlib format."""
        # precursor_mz: Union[List[float], np.ndarray],
        # precursor_charges: Union[List[int], np.ndarray],
        # modified_sequences: List[str],
        # retention_times: Union[List[float], np.ndarray],
        # fragmentmz: List[np.ndarray],
        # intensities: List[np.ndarray],

        intensities = self.grpc_output[list(self.grpc_output)[0]]["intensity"]
        fragment_mz = self.grpc_output[list(self.grpc_output)[0]]["fragmentmz"]
        # annotation = self.grpc_output[list(self.grpc_output)[0]]["annotation"]
        irt = self.grpc_output[list(self.grpc_output)[1]]
        retention_times = irt.flatten()
        modified_sequences = self.spectra_input["MODIFIED_SEQUENCE"]

        precursor_charges = self.spectra_input["PRECURSOR_CHARGE"]
        precursor_masses = self.spectra_input["MASS"]
        precursor_mz = (precursor_masses + (precursor_charges * PARTICLE_MASSES["PROTON"])) / precursor_charges

        self.create_database(self.out_path)

        # gather all values for the entries table and create pandas DataFrame
        masked_values = self._calculate_masked_values(fragment_mz, intensities)
        mass_mod_sequences = internal_to_mod_mass(modified_sequences)
        sequences = internal_without_mods(modified_sequences)
        data_list = [*masked_values, precursor_charges, mass_mod_sequences, sequences, retention_times, precursor_mz]
        self.entries = pd.DataFrame(dict(zip(DLIB_COL_NAMES, data_list)))

        # hardcoded entries that we currently not use.
        # Visit https://bitbucket.org/searleb/encyclopedia/wiki/EncyclopeDIA%20File%20Formats for dlib specs
        self.entries["Copies"] = 1  # this is hardcorded for now and unused
        self.entries["Score"] = 0
        self.entries["CorrelationEncodedLength"] = None
        self.entries["CorrelationArray"] = None
        self.entries["RTInSecondsStart"] = None
        self.entries["RTInSecondsStop"] = None
        self.entries["MedianChromatogramEncodedLength"] = None
        self.entries["MedianChromatogramArray"] = None
        self.entries["SourceFile"] = "Prosit"

        # gather all values for the p2p table and create pandas DataFrame
        self.p2p = pd.DataFrame({"PeptideSeq": sequences, "isDecoy": False, "ProteinAccession": "unknown"})



from math import ceil
from typing import TypeVar

from .peak import Peak

# import pyximport; pyximport.install()
# from mgf_filter.cython.mat import ceil


MasterPeakT = TypeVar("MasterPeakT", bound="MasterPeak")


class MasterPeak(Peak):
    """Container class for a aggregated, summed up fragment peak."""

    def __init__(self, peak: Peak):
        """
        Contructor for a master peak.

        # TODO provide details

        :param peak: a peak object # TODO provide details
        """
        # right and counts are calculated based on update func
        self.right = 0
        self.counts = 1
        self.rel_intensity_ratio = 0.0
        # ratio is set to an actual value first time by comparing a spectrum to another spectrum
        self.counts_ratio = 0.0
        self.mz_origin = peak.mz
        super().__init__(peak.mz, peak.intensity, peak.delta_function, meta=peak.meta)
        # update does not have to be called, because constructor
        # of peak calls update of MasterPeak
        # self.update()

    def __eq__(self, other: MasterPeakT) -> bool:  # type: ignore
        """
        Reports true if both master peaks have the same member variables!

        :param other: the other master peak object to compare this object with
        :return: whether or not the two objects are equal
        """
        if self.__dict__ == other.__dict__:
            return True
        else:
            if (
                (self.counts == other.counts)
                & (self.mz == other.mz)
                & (self.left == other.left)
                & (self.right == other.right)
                & (self.mz_origin == other.mz_origin)
            ):
                return True
            else:
                return False

    def __ne__(self, other: MasterPeakT) -> bool:  # type: ignore
        """
        Reports true if any member variable differs between the two master peaks!

        :param other: the other master peak object to compare this object with
        :return: whether or not the two objects are equal
        """
        if self.__dict__ != other.__dict__:
            if (
                (self.counts == other.counts)
                & (self.mz == other.mz)
                & (self.left == other.left)
                & (self.right == other.right)
                & (self.mz_origin == other.mz_origin)
            ):
                return False
            else:
                return True
        else:
            return True  # TODO check if this is really true, dicts equal means the objects should be equal

    def __str__(self) -> str:
        """
        Overrides the default str method to provide information about member variables.

        :return: string representation of this object
        """
        return (
            "mz: "
            + str(self.mz)
            + "\n"
            + "intensity: "
            + str(self.intensity)
            + "\n"
            + "left: "
            + str(self.left)
            + "\n"
            + "right: "
            + str(self.right)
            + "\n"
            + "counts: "
            + str(self.counts)
            + "\n"
            + "rel_intensity_ratio: "
            + str(self.rel_intensity_ratio)
            + "\n"
            + "counts_ratio: "
            + str(self.counts_ratio)
            + "\n"
            + "origin: "
            + str(self.mz_origin)
        )

    def update(self):
        """Calculates delta, left and right."""
        self.delta = self.delta_function(self.mz)
        self.left = self.mz - self.delta
        self.right = self.mz + self.delta
        self.ceiled_key = ceil(self.left)

    def is_inside(self, peak: Peak) -> bool:
        """
        Reports true if tested peak is inside window.

        :param peak: a peak object to check

        :return: whether the peak's mz is within left and right border
        """
        return (self.left < peak.mz) and (self.right > peak.mz)

    def is_inside_mz(self, mz: float) -> bool:
        """
        Reports true if tested mz is inside window.

        :param mz: the mz to check for

        :return: whether mz is within left and right border
        """
        return (self.left < mz) and (self.right > mz)

    def add(self, peak: Peak):
        """
        Default_counts must be overwritten if more than one peak is added (multimerge).

        :param peak: a peak object to be added to this master peak.
        """
        self.mz = (self.mz * self.intensity + peak.mz * peak.intensity) / (peak.intensity + self.intensity)
        self.intensity = self.intensity + peak.intensity
        self.counts += peak.counts
        self.update()

    def smaller(self, peak: Peak) -> bool:
        """
        Report true if MasterPeak window is below tested peak  # TODO make this a magic function.

        :param peak: the other peak to compare to
        :return: wheter or not the peak is smaller to this object
        """
        return self.right < peak.mz

    def greater(self, peak: Peak):
        """
        Report true if MasterPeak window is above tested peak  # TODO make this a magic function.

        :param peak: the other peak to compare to
        :return: wheter or not the peak is greater to this object
        """
        return self.left > peak.mz

    def recalculate_ratio(self, mp: MasterPeakT):
        """
        Recalculate the ratio  # TODO details.

        :param mp: the other master peak object
        """
        rel_ratio = (self.intensity / self.counts) / (mp.intensity / mp.counts)
        self.rel_intensity_ratio = rel_ratio
        rel_counts_ratio = self.counts / mp.counts
        self.counts_ratio = rel_counts_ratio



import logging
from pathlib import Path
from typing import Union

import pandas as pd
import spectrum_fundamentals.constants as c
from spectrum_fundamentals.mod_string import sage_to_internal

from .search_results import SearchResults, filter_valid_prosit_sequences

logger = logging.getLogger(__name__)


class Sage(SearchResults):
    """Handle search results from Sage."""

    def read_result(self, tmt_labeled: str = "") -> pd.DataFrame:
        """
        Function to read a msms tsv and perform some basic formatting.

        :param tmt_labeled: tmt label as str
        :return: pd.DataFrame with the formatted data
        """
        logger.info("Reading msms.tsv file")
        df = pd.read_csv(
            self.path,
            usecols=["filename", "scannr", "peptide", "charge", "hyperscore", "calcmass", "proteins"],
            sep="\t",
        )
        logger.info("Finished reading msms.tsv file")

        # Standardize column names
        df.columns = df.columns.str.upper()
        df.columns = df.columns.str.replace(" ", "_")

        df = Sage.update_columns_for_prosit(df, tmt_labeled)
        return filter_valid_prosit_sequences(df)

    @staticmethod
    def update_columns_for_prosit(df: pd.DataFrame, tmt_labeled: str) -> pd.DataFrame:
        """
        Update columns of df to work with Prosit.

        :param df: df to modify
        :param tmt_labeled: True if tmt labeled, ignored
        :return: modified df as pd.DataFrame
        """
        df = df.rename(
            columns={
                "FILENAME": "RAW_FILE",
                "SCANNR": "SCAN_NUMBER",
                "PEPTIDE": "MODIFIED_SEQUENCE",
                "CHARGE": "PRECURSOR_CHARGE",
            }
        )

        # removing .mzML
        df["RAW_FILE"] = df["RAW_FILE"].str.replace(".mzML", "", regex=True)
        # extracting only the scan number
        df["SCAN_NUMBER"] = [int(x.rsplit("=", 1)[-1]) for x in df["SCAN_NUMBER"]]
        # creating a column of decoys and targets
        df["REVERSE"] = df["PROTEINS"].str.startswith("rev_")
        # removing modification to create the unmodified sequences
        df["SEQUENCE"] = df["MODIFIED_SEQUENCE"].str.replace(r"\[.*?\]", "", regex=True)
        # length of the peptide
        df["PEPTIDE_LENGTH"] = df["SEQUENCE"].str.len()
        # mass of the peptide
        df["MASS"] = df["CALCMASS"]
        # score of the peptide
        df["SCORE"] = df["HYPERSCORE"]
        # converting proforma to unimode
        print(df)
        df["MODIFIED_SEQUENCE"] = sage_to_internal(df["MODIFIED_SEQUENCE"])

        print(df.columns)
        return df



import logging
from pathlib import Path
from typing import Union

import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES

from .search_results import filter_valid_prosit_sequences

logger = logging.getLogger(__name__)


def _update_columns_for_prosit(df: pd.DataFrame):
    """
    Update columns of df to work with Prosit.

    :param df: df to modify
    :return: modified df as pd.DataFrame
    """

    def _check_unsupported_mods(modstring: str):
        if modstring:
            for mod in modstring.split(";"):
                if not mod.split("(", 1)[1].startswith(("Oxidation", "Carbamidomethyl")):
                    raise AssertionError(f"Unsupported modification: {mod}")

    def _make_mod_seq(seq: str):
        seq = seq.replace("m", "M[UNIMOD:35]")
        seq = seq.replace("c", "C[UNIMOD:4]")
        return seq

    df["REVERSE"] = df["Protein Accessions"].str.startswith("REV_")
    df["MASS"] = df["m/z"] * df["PRECURSOR_CHARGE"] - (df["PRECURSOR_CHARGE"] * PARTICLE_MASSES["PROTON"])
    df["Modifications"].fillna("").apply(_check_unsupported_mods)
    df["MODIFIED_SEQUENCE"] = df["SEQUENCE"].apply(_make_mod_seq)
    df["SEQUENCE"] = df["SEQUENCE"].str.upper()
    df["PEPTIDE_LENGTH"] = df["SEQUENCE"].str.len()
    df["RAW_FILE"] = df["RAW_FILE"].str.replace(r"\.\w+$", "", regex=True)

    return df[
        [
            "SCORE",
            "MASS",
            "PRECURSOR_CHARGE",
            "RAW_FILE",
            "SCAN_NUMBER",
            "REVERSE",
            "MODIFIED_SEQUENCE",
            "SEQUENCE",
            "PEPTIDE_LENGTH",
        ]
    ]


def _remove_decoys_in_targets(full_df):
    duplicated_psms = full_df[["SCAN_NUMBER", "RAW_FILE", "MODIFIED_SEQUENCE"]].duplicated(keep=False)
    logger.info(f"Removing {sum(duplicated_psms)} duplicated PSMs...")
    full_df = full_df[~(full_df["REVERSE"] & duplicated_psms)]
    if any(full_df[["SCAN_NUMBER", "RAW_FILE", "MODIFIED_SEQUENCE"]].duplicated(keep=False)):
        raise AssertionError("There are duplicate target PSMs. This is a bug of msamanda!")
    logger.info(f"{len(full_df)} remaining PSMs")

    return full_df


def read_result(path: Union[str, Path], suffix: str = "output.csv") -> pd.DataFrame:
    """
    Function to read a msms txt and perform some basic formatting.

    :param path: path to msms.txt to read
    :param suffix: Optional suffix to determine which fileresult files should be taken from the supplied path
    :raises FileNotFoundError: If the supplied path is not found
    :raises AssertionError: If the supplied path does not contain any files matching the provided suffix.
    :return: pd.DataFrame with the formatted data
    """
    if isinstance(path, str):
        path = Path(path)
    if path.is_file():
        pathlist = [path]
    elif path.is_dir():
        pathlist = list(path.glob(f"*{suffix}"))
        if not pathlist:
            raise AssertionError(f"The directory does not contain any files that match the pattern *{suffix}")
    else:
        raise FileNotFoundError(f"{path} does not exist.")

    df_list = []
    for output_file in pathlist:
        logger.info(f"Reading {output_file}...")
        df = pd.read_csv(
            output_file,
            sep="\t",
            skiprows=1,
            usecols=[
                "Scan Number",
                "Sequence",
                "Modifications",
                "Filename",
                "Amanda Score",
                "m/z",
                "Charge",
                "Protein Accessions",
            ],
        )
        df.columns = [
            "SCAN_NUMBER",
            "SEQUENCE",
            "Modifications",
            "Protein Accessions",
            "SCORE",
            "m/z",
            "PRECURSOR_CHARGE",
            "RAW_FILE",
        ]
        logger.info(f"Finished reading {output_file}")

        df = _update_columns_for_prosit(df)
        df = filter_valid_prosit_sequences(df)
        df_list.append(df)

    if len(df_list) == 1:
        full_df = df_list[0]
    full_df = pd.concat(df_list, axis=0)
    full_df = _remove_decoys_in_targets(full_df)
    return full_df



import csv
from math import ceil
from typing import Callable, List, Optional

from sortedcontainers import SortedDict, SortedList

from .masterPeak import MasterPeak
from .peak import Peak

"""
def merge_func(offset: int, idx: int):
    master_peak_bordering = self.spectrum[charge][key][idx + offset]
    get_idx_master_peak = self.spectrum[charge][key][idx]
    # left idx -1 : idx +1  # right idx : idx+2
    # left offset = -1 -> +2 -1 = 1
    # right offset = 0 -> +2 +0 = 2
    del self.spectrum[charge][key][
        idx + offset : idx + 2 + offset
    ]  # delete is slicing, therefore one more
    if len(self.spectrum[charge][key]) == 0:
        del self.spectrum[charge][key]
    get_idx_master_peak.add(master_peak_bordering)
    get_idx_master_peak.add(peak)
    self.multimerged += 1
    if get_idx_master_peak.key() not in self.spectrum[charge]:
        self.spectrum[charge][get_idx_master_peak.key()] = SortedList(key=lambda i: i.left)
    self.spectrum[charge][get_idx_master_peak.key()].add(get_idx_master_peak)
"""


def _calculate_delta_by_ppm(ppm: int) -> Callable:
    def fix_ppm(mz: float):
        return ppm * float(mz) / (pow(10, 6))

    return fix_ppm


def _calculate_relative_intensity(a_intensity: List[int]) -> List[float]:
    max_intensity = max(a_intensity)
    return [x / max_intensity for x in a_intensity]


class MasterSpectrum:
    """
    Container class for a Spectrum.

    Master spectrum takes peaks and converts them to master peaks. If there are no master peaks
    to be inserted spectrum is {charge of spectrum: {bins: MP}}.
    """

    def __init__(self):
        """Constructor for MasterSpectrum."""
        self.spectrum = {}
        self.merged = 0
        self.appended = 0
        self.multimerged = 0  # within one insertion 3 peaks are merged

    def binary(self, peak: Peak, imin: int, imax: int, charge: int):
        """
        Binary search.

        :param peak: #TODO
        :param imin: the minimum search position
        :param imax: the max search position
        :param charge: defines in which masterspectrum to search
        :return: tuple with the following members  # TODO proper documentation:
            position of insert, -1 if can not be added to any peak
            add peak from left or right bin
            must left peak also be added (merge case)
            right peak also be added (think about 3 peaks and a merge between 2 and 3)
        """
        key = peak.key()
        imid = int(ceil((imax + imin) / 2))
        if imax < imin:
            exist_left_bin = key - 1 in self.spectrum[charge].keys()
            exist_right_bin = key + 1 in self.spectrum[charge].keys()

            if exist_left_bin and self.spectrum[charge][key - 1][-1].is_inside(peak):
                return -1, -1, False, False
            if exist_right_bin and self.spectrum[charge][key + 1][0].is_inside(peak):
                return -1, 1, False, False
            return -1, 0, False, False

        # search must go on
        mspeak_in_bin = self.spectrum[charge][key][imid]
        if mspeak_in_bin.greater(peak):
            return self.binary(peak, imin, imid - 1, charge)
        elif mspeak_in_bin.smaller(peak):
            return self.binary(peak, imid + 1, imax, charge)
        # search results in peak that should be added
        else:
            if imid == 0:
                exist_left_bin = key - 1 in self.spectrum[charge]
                exist_right_bin = key + 1 in self.spectrum[charge]

                if exist_left_bin and self.spectrum[charge][key - 1][-1].is_inside(peak):
                    return 0, -1, False, False
                elif exist_right_bin and self.spectrum[charge][key + 1][0].is_inside(peak):
                    return 0, 1, False, False
                return 0, 0, False, False

            else:  # peak is somewhere between 1 and last
                # imid -1 exists alway
                is_last_entry = len(self.spectrum[charge][key]) - 1 == imid
                exists_bigger_peak = not (is_last_entry)
                if exists_bigger_peak:
                    if self.spectrum[charge][key][imid - 1].is_inside(peak):
                        return imid, 0, True, False
                    if self.spectrum[charge][key][imid + 1].is_inside(peak):
                        return imid, 0, False, True
                    return imid, 0, False, False
                else:  # is last entry
                    exist_right_bin = key + 1 in self.spectrum[charge]
                    if self.spectrum[charge][key][imid - 1].is_inside(peak):
                        return imid, 0, True, False
                    elif exist_right_bin and self.spectrum[charge][key + 1][0].is_inside(peak):
                        return imid, 1, False, False
                    return imid, 0, False, False

    def add(self, peak, charge: int = 0):
        """
        Add charges to peaks.

        Charge is by default 0, so if a spectrum should be summed without including charge information
        it is defaulted to 0.

        :param peak: TODO
        :param charge: TODO
        """
        key = peak.key()
        if charge not in self.spectrum:
            self.spectrum[charge] = SortedDict()

        if key in self.spectrum[charge]:
            idx, bin_to_ack, should_merge_left_peak, should_merge_right_peak = self.binary(
                peak, 0, len(self.spectrum[charge][key]) - 1, charge
            )
            if idx == -1:  # does not have to react to merge cases!
                if bin_to_ack == 0:
                    self.appended += 1
                    if type(peak).__name__ == "Peak":
                        self.spectrum[charge][key].add(MasterPeak(peak))
                    else:
                        self.spectrum[charge][key].add(peak)
                elif bin_to_ack == -1:
                    if len(self.spectrum[charge][key]) == 0:
                        del self.spectrum[charge][key]
                    get_master_peak_left_bin = self.spectrum[charge][key - 1][-1]
                    del self.spectrum[charge][key - 1][-1]
                    if len(self.spectrum[charge][key - 1]) == 0:
                        del self.spectrum[charge][key - 1]
                    get_master_peak_left_bin.add(peak)
                    self.merged += 1
                    if get_master_peak_left_bin.key() not in self.spectrum[charge]:
                        self.spectrum[charge][get_master_peak_left_bin.key()] = SortedList(key=lambda i: i.left)
                    self.spectrum[charge][get_master_peak_left_bin.key()].add(get_master_peak_left_bin)
                else:  # +1
                    if len(self.spectrum[charge][key]) == 0:
                        del self.spectrum[charge][key]
                    get_master_peak_right_bin = self.spectrum[charge][key + 1][0]
                    del self.spectrum[charge][key + 1][0]
                    if len(self.spectrum[charge][key + 1]) == 0:
                        del self.spectrum[charge][key + 1]
                    get_master_peak_right_bin.add(peak)
                    self.merged += 1
                    if get_master_peak_right_bin.key() not in self.spectrum[charge]:
                        self.spectrum[charge][get_master_peak_right_bin.key()] = SortedList(key=lambda i: i.left)
                    self.spectrum[charge][get_master_peak_right_bin.key()].add(get_master_peak_right_bin)

            else:  # idx != -1
                if bin_to_ack == 0:
                    if should_merge_left_peak:
                        get_left_master_peak = self.spectrum[charge][key][idx - 1]
                        get_idx_master_peak = self.spectrum[charge][key][idx]
                        del self.spectrum[charge][key][idx - 1 : idx + 1]  # delete is slicing, therefore one more
                        if len(self.spectrum[charge][key]) == 0:
                            del self.spectrum[charge][key]
                        get_idx_master_peak.add(get_left_master_peak)
                        get_idx_master_peak.add(peak)
                        self.multimerged += 1
                        if get_idx_master_peak.key() not in self.spectrum[charge]:
                            self.spectrum[charge][get_idx_master_peak.key()] = SortedList(key=lambda i: i.left)
                        self.spectrum[charge][get_idx_master_peak.key()].add(get_idx_master_peak)
                    elif should_merge_right_peak:
                        get_idx_master_peak = self.spectrum[charge][key][idx]
                        get_right_master_peak = self.spectrum[charge][key][idx + 1]
                        del self.spectrum[charge][key][idx : idx + 2]  # delete is slicing, therefore one more
                        if len(self.spectrum[charge][key]) == 0:
                            del self.spectrum[charge][key]
                        get_idx_master_peak.add(get_right_master_peak)
                        get_idx_master_peak.add(peak)
                        self.multimerged += 1
                        if get_idx_master_peak.key() not in self.spectrum[charge]:
                            self.spectrum[charge][get_idx_master_peak.key()] = SortedList(key=lambda i: i.left)
                        self.spectrum[charge][get_idx_master_peak.key()].add(get_idx_master_peak)
                    else:  # case idx = 0, 0, False, False
                        get_idx_master_peak = self.spectrum[charge][key][idx]
                        del self.spectrum[charge][key][idx]  # delete is slicing, therefore one more
                        if len(self.spectrum[charge][key]) == 0:
                            del self.spectrum[charge][key]
                        get_idx_master_peak.add(peak)
                        self.merged += 1
                        if get_idx_master_peak.key() not in self.spectrum[charge]:
                            self.spectrum[charge][get_idx_master_peak.key()] = SortedList(key=lambda i: i.left)
                        self.spectrum[charge][get_idx_master_peak.key()].add(get_idx_master_peak)
                elif bin_to_ack == -1:  # idx != -1
                    # 0, -1, F, F
                    get_idx_master_peak = self.spectrum[charge][key][idx]
                    get_master_peak_before = self.spectrum[charge][key - 1][-1]
                    del self.spectrum[charge][key][idx]
                    if len(self.spectrum[charge][key]) == 0:
                        del self.spectrum[charge][key]
                    del self.spectrum[charge][key - 1][-1]
                    if len(self.spectrum[charge][key - 1]) == 0:
                        del self.spectrum[charge][key - 1]
                    get_idx_master_peak.add(get_master_peak_before)
                    get_idx_master_peak.add(peak)
                    self.merged += 1
                    if get_idx_master_peak.key() not in self.spectrum[charge]:
                        self.spectrum[charge][get_idx_master_peak.key()] = SortedList(key=lambda i: i.left)
                    self.spectrum[charge][get_idx_master_peak.key()].add(get_idx_master_peak)
                else:  # bin_to_ack == 1, idx !=  -1
                    get_idx_master_peak = self.spectrum[charge][key][idx]
                    get_master_peak_after = self.spectrum[charge][key + 1][0]
                    del self.spectrum[charge][key][idx]
                    if len(self.spectrum[charge][key]) == 0:
                        del self.spectrum[charge][key]
                    del self.spectrum[charge][key + 1][0]
                    if len(self.spectrum[charge][key + 1]) == 0:
                        del self.spectrum[charge][key + 1]
                    get_idx_master_peak.add(get_master_peak_after)
                    get_idx_master_peak.add(peak)
                    self.merged += 1
                    if get_idx_master_peak.key() not in self.spectrum[charge]:
                        self.spectrum[charge][get_idx_master_peak.key()] = SortedList(key=lambda i: i.left)
                    self.spectrum[charge][get_idx_master_peak.key()].add(get_idx_master_peak)
        else:
            self.spectrum[charge][key] = SortedList(key=lambda i: i.left)
            self.add(peak, charge)

    def load_from_tims(
        self, intensities: List[int], mzs: List[float], ignore_charges: bool, delta_func: Optional[Callable] = None
    ):
        """
        Load data from tims and create master spectrum.

        :param intensities: list of intensities of peaks of individual spectra to sum
        :param mzs: list of mzs of peaks of individual spectra to sum
        :param ignore_charges: whether to ignore charges when summing up peaks
        :param delta_func: a callable to calculate the mass tolerance window
        :raises NotImplementedError: If ignore_charges is set to False
        """
        rel_int = _calculate_relative_intensity(intensities)
        if delta_func is None:
            delta_func = _calculate_delta_by_ppm(40)
        for m, i in zip(mzs, rel_int):
            p = Peak(float(m), float(i), delta_func)
            if ignore_charges:
                self.add(p, 0)
            else:
                raise NotImplementedError("Adding up intensities using precursor charge is not supported.")



from typing import IO, Dict

import numpy as np
import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES
from spectrum_fundamentals.mod_string import internal_to_mod_names, internal_without_mods

from .spectral_library import SpectralLibrary


class MSP(SpectralLibrary):
    """Main to initialze a MSP obj."""

    @staticmethod
    def _assemble_fragment_string(f_mz: float, f_int: float, f_a: bytes):
        annot = f_a[:-2].decode() if f_a.endswith(b"1") else f_a.replace(b"+", b"^").decode()
        return f'{f_mz:.8f}\t{f_int:.4f}\t"{annot}/0.0ppm"\n'

    def _write(self, out: IO, data: Dict[str, np.ndarray], metadata: pd.DataFrame):
        # prepare metadata
        stripped_peptides = metadata["SEQUENCE"]
        modss = internal_to_mod_names(metadata["MODIFIED_SEQUENCE"])
        p_charges = metadata["PRECURSOR_CHARGE"]
        p_mzs = (metadata["MASS"] + (p_charges * PARTICLE_MASSES["PROTON"])) / p_charges
        ces = metadata["COLLISION_ENERGY"]

        # prepare spectra
        irts = data["irt"][:, 0]  # should create a 1D view of the (n_peptides, 1) shaped array
        f_mzss = data["mz"]
        f_intss = data["intensities"]
        f_annotss = data["annotation"]

        lines = []
        vec_assemble = np.vectorize(MSP._assemble_fragment_string)

        for stripped_peptide, p_charge, p_mz, ce, mods, irt, f_mzs, f_ints, f_annots in zip(
            stripped_peptides, p_charges, p_mzs, ces, modss, irts, f_mzss, f_intss, f_annotss
        ):
            lines.append(f"Name: {stripped_peptide}/{p_charge}\nMW: {p_mz}\n")
            lines.append(
                f"Comment: Parent={p_mz:.8f} Collision_energy={ce} Mods={mods[0]} "
                f"ModString={mods[1]}/{p_charge} iRT={irt:.2f}\n"
            )

            cond = self._fragment_filter_passed(f_mzs, f_ints)
            fragment_list = vec_assemble(f_mzs[cond], f_ints[cond], f_annots[cond])

            lines.append(f"Num peaks: {len(fragment_list)}\n")

            lines.extend(fragment_list)
        out.writelines(lines)

    def _write_header(self, out: IO):
        pass



import argparse
import collections
import csv
import itertools
import logging
import sys
from typing import Dict, List

import numpy as np

logger = logging.getLogger(__name__)

cleavage_sites = {
    "trypsinp": (["K", "R"], []),
    "trypsin": (["K", "R"], ["P"]),
    "chymotrypsin": (["F", "W", "Y", "L"], ["P"]),
    "proteinasek": (["A", "E", "F", "I", "L", "T", "V", "W", "Y"], []),
    "elastase": (["L", "V", "A", "G"], ["P"]),
    "lys-c": (["K"], ["P"]),
    "arg-c": (["R"], ["P"]),
    "glu-c": (["E"], ["P"]),
    "v8-de": (["N", "D", "E", "Q"], ["P"]),
    "no_enzyme": ([], []),
}


def main(args):
    """Main to run digestion."""
    args = parse_args(args)
    # python digest.py --enzyme trypsinp --cleavages 0 --fasta /home/matthewt/data/MaxLFQ_benchmark/ups.fasta
    # --prosit_input ~/data/Prosit_test/ups_prosit_input.csv
    if args.prosit_input:
        writer = get_tsv_writer(args.prosit_input, delimiter=",")
        writer.writerow("modified_sequence,collision_energy,precursor_charge,fragmentation".split(","))
        prosit_input_file_with_proteins = args.prosit_input.replace(".csv", "_with_proteins.csv")
        writer_with_proteins = get_tsv_writer(prosit_input_file_with_proteins, delimiter=",")
        writer_with_proteins.writerow(
            "modified_sequence,collision_energy,precursor_charge,protein,fragmentation".split(",")
        )

        for peptide, proteins in get_peptide_to_protein_map(
            args.fasta,
            db=args.db,
            digestion=args.digestion,
            min_len=args.min_length,
            max_len=args.max_length,
            enzyme=args.enzyme,
            miscleavages=args.cleavages,
            methionine_cleavage=True,
            special_aas=list(args.special_aas),
            use_hash_key=False,
        ).items():
            if valid_prosit_peptide(peptide):
                for charge in [2, 3, 4]:
                    writer.writerow([peptide, 30, charge, args.fragmentation])
                    writer_with_proteins.writerow([peptide, 30, charge, proteins[0], args.fragmentation])

    if args.peptide_protein_map:
        with open(args.peptide_protein_map + ".params.txt", "w") as f:
            f.write(" ".join(sys.argv))
        writer = get_tsv_writer(args.peptide_protein_map, delimiter="\t")

        for peptide, proteins in get_peptide_to_protein_map(
            args.fasta,
            db="concat",
            digestion=args.digestion,
            min_len=args.min_length,
            max_len=args.max_length,
            enzyme=args.enzyme,
            miscleavages=args.cleavages,
            methionine_cleavage=True,
            special_aas=list(args.special_aas),
            use_hash_key=False,
        ).items():
            writer.writerow([peptide, ";".join(proteins)])

    if args.ibaq_map:
        writer = get_tsv_writer(args.ibaq_map, delimiter="\t")

        num_peptides_per_protein = get_num_ibaq_peptides_per_protein(args)

        for protein, num_peptides in num_peptides_per_protein.items():
            writer.writerow([protein, num_peptides])


def valid_prosit_peptide(peptide: str) -> bool:
    """Check if peptide is valid."""
    return len(peptide) <= 30 and "U" not in peptide and "X" not in peptide


def parse_args(call):
    """Parse args."""
    apars = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    apars.add_argument(
        "--fasta",
        metavar="F",
        required=False,
        help="""Fasta file used as input
                                                    """,
    )

    apars.add_argument(
        "--prosit_input",
        metavar="M",
        required=False,
        help="""Path to file where to write the prosit input file.
                                                    """,
    )

    apars.add_argument(
        "--peptide_protein_map",
        default=None,
        metavar="M",
        required=False,
        help="""Write mapping from peptides to all its proteins to
                                                         the specified file.
                                                    """,
    )

    apars.add_argument(
        "--ibaq_map",
        default=None,
        metavar="M",
        required=False,
        help="""Write number of peptides per protein to the specified
                                                         file that meet the iBAQ criteria
                                                         (6 <= pepLen <= 30, no miscleavages).
                                                    """,
    )
    apars.add_argument(
        "--fragmentation", default=None, metavar="M", required=True, help="""Fragmentation (HCD or CID)"""
    )

    apars.add_argument("--db", metavar="M", required=False, default="concat", help="""Target/decoy/concat""")
    add_arguments(apars)
    args = apars.parse_args(call)
    return args


def add_arguments(apars):
    """Add arguments to arg parser."""
    apars.add_argument(
        "-e",
        "--enzyme",
        default="trypsin",
        metavar="E",
        help="""Type of enzyme "no_enzyme","elastase","pepsin",
                                                         "proteinasek","thermolysin","chymotrypsin",
                                                         "lys-n","lys-c","arg-c","asp-n","glu-c","trypsin",
                                                         "trypsinp".
                                                    """,
    )

    apars.add_argument(
        "-c",
        "--cleavages",
        default=2,
        metavar="C",
        type=int,
        help="""Number of allowed miss cleavages used in the search
                                                         engine (Only valid when using option -F).
                                                    """,
    )

    apars.add_argument(
        "-l",
        "--min-length",
        default=7,
        metavar="L",
        type=int,
        help="""Minimum peptide length allowed used in the search
                                                         engine (Only valid when using option -F).
                                                    """,
    )

    apars.add_argument(
        "-t",
        "--max-length",
        default=60,
        metavar="L",
        type=int,
        help="""Maximum peptide length allowed used in the search
                                                         engine (Only valid when using option -F).
                                                    """,
    )

    apars.add_argument(
        "--special-aas",
        default="KR",
        metavar="S",
        help="""Special AAs that MaxQuant uses for decoy generation.
                                                    """,
    )

    apars.add_argument(
        "--digestion",
        default="full",
        metavar="D",
        help="""Digestion mode ('full', 'semi' or 'none').
                                                    """,
    )


def write_protein_to_gene_map(fasta_file, output_file):
    """Write protein to gene map."""
    writer = csv.writer(open(output_file, "w"), delimiter="\t")
    for protein_name, _ in read_fasta_tide(fasta_file, db="target"):
        protein_id = protein_name.split("|")[1]
        gene_id = protein_name.split("|")[2].split(" ")[0]
        writer.writerow([protein_id, gene_id])


parse_until_first_space = lambda x: x.split(" ")[0]
parseprotein_name_func = lambda x: " ".join(x.split(" OS=")[0].split(" ")[1:])
parse_gene_name_func = lambda x: x.split(" GN=")[1].split(" ")[0] if "GN=" in x else ""


def parse_uniprot_id(fasta_id):
    """Parse uniprot id."""
    protein_id = parse_until_first_space(fasta_id)
    if "|" in protein_id:
        return protein_id.split("|")[1]
    else:
        return protein_id


def read_fasta_proteins(
    file_path,
    db="concat",
    parse_id=parse_until_first_space,
    parseprotein_name=parseprotein_name_func,
    parse_gene_name=parse_gene_name_func,
):
    """Read fasta proteins."""
    with open(file_path) as fp:
        for line in itertools.chain(fp, [">"]):
            line = line.rstrip()
            if line.startswith(">"):
                if db in ["target", "concat"]:
                    yield (parse_id(line[1:]), parseprotein_name(line[1:]), parse_gene_name(line[1:]), line[1:])

                if db in ["decoy", "concat"]:
                    yield ("REV__" + parse_id(line[1:]), "", "REV__" + parse_gene_name(line[1:]), "")


def get_protein_annotations(fasta_file, parse_id):
    """Get protein annotations."""
    protein_annotations = dict()
    if not fasta_file:
        return protein_annotations

    for protein, protein_name, gen_name, fasta_header in read_fasta_proteins(fasta_file, parse_id=parse_id):
        if protein not in protein_annotations:
            protein_annotations[protein] = (protein_name, gen_name, fasta_header)

    return protein_annotations


def has_gene_names(protein_annotations) -> bool:
    """Check if gene names are provided."""
    counts = sum(1 for _, gen_name, _ in protein_annotations.values() if len(gen_name) > 0)
    return counts / len(protein_annotations) > 0.5


def read_fasta_tide(file_path, db: str = "target", parse_id=parse_until_first_space):
    """Read fasta tide."""
    read_fasta_maxquant(file_path, db, parse_id, special_aas=[], decoy_prefix="decoy_")


def read_fasta_maxquant(
    file_path, db: str = "target", parse_id=parse_until_first_space, special_aas=None, decoy_prefix: str = "REV__"
):
    """Read fasta maxquant."""
    if special_aas is None:
        special_aas = ["K", "R"]
    if db not in ["target", "decoy", "concat"]:
        sys.exit("unknown db mode: %s" % db)

    hasspecial_aas = len(special_aas) > 0
    with open(file_path) as fp:
        line = next(fp).rstrip()
        name = parse_id(line[1:])
        sequence_lines: List[str] = []
        for line in itertools.chain(fp, [">"]):
            line = line.rstrip()
            if line.startswith(">"):
                sequence = "".join(sequence_lines)
                if db in ["target", "concat"]:
                    yield name, sequence
                if db in ["decoy", "concat"]:
                    rev_sequence = sequence[::-1]
                    if hasspecial_aas:
                        rev_sequence = swap_special_aas(rev_sequence, special_aas)
                    yield decoy_prefix + name, rev_sequence
                name = parse_id(line[1:])
                sequence_lines = []
                continue
            sequence_lines.append(line)


read_fasta = read_fasta_maxquant


# e.g. special_aas = ['R', 'K'] transforms ABCKDEFRK into ABKCDERKF
def swap_special_aas(seq, special_aas):
    """Swaps the special AAs with its preceding amino acid, as is done in MaxQuant."""
    seq = list(seq)
    for i in range(1, len(seq)):
        if seq[i] in special_aas:
            swap_positions(seq, i, i - 1)
    seq = "".join(seq)
    return seq


def swap_positions(seq, pos1, pos2):
    """Swaps positions."""
    seq[pos1], seq[pos2] = seq[pos2], seq[pos1]


def get_protein_ids(file_path):
    """Get protein ids."""
    protein_ids = list()
    for protein_id, _ in read_fasta(file_path):
        protein_ids.append(protein_id)
    return set(protein_ids)


def get_protein_sequences(file_path, parse_id):
    """Get protein sequences."""
    protein_sequences = dict()
    for protein_id, protein_seq in read_fasta(file_path, db="concat", parse_id=parse_id):
        if protein_id not in protein_sequences:  # keep only first sequence per identifier
            protein_sequences[protein_id] = protein_seq
    return protein_sequences


def filter_fasta_file(fasta_file, filteredfasta_file, proteins):
    """Filter fasta file."""
    with open(filteredfasta_file, "w") as f:
        for prot, seq in read_fasta(fasta_file):
            if prot in proteins:
                f.write(">" + prot + "\n" + seq + "\n")


def get_peptides(
    fasta_file,
    db="concat",
    min_len=6,
    max_len=50,
    pre=None,
    not_post=None,
    digestion="full",
    miscleavages=0,
    methionine_cleavage=True,
):
    """Get peptides."""
    if pre is None:
        pre = ["K", "R"]
    if not_post is None:
        not_post = ["P"]
    for _, seq in read_fasta(fasta_file, db):
        yield from get_digested_peptides(
            seq, min_len, max_len, pre, not_post, digestion, miscleavages, methionine_cleavage
        )


def get_digested_peptides(
    seq,
    min_len=6,
    max_len=50,
    pre=None,
    not_post=None,
    digestion="full",
    miscleavages=0,
    methionine_cleavage=True,
):
    """Get digested peptides."""
    if pre is None:
        pre = ["K", "R"]
    if not_post is None:
        not_post = ["P"]
    if digestion == "none":
        yield from non_specific_digest(seq, min_len, max_len)
    elif digestion == "semi":
        yield from semi_specific_digest(seq, min_len, max_len, pre, not_post, miscleavages, methionine_cleavage)
    else:
        yield from full_digest(seq, min_len, max_len, pre, not_post, miscleavages, methionine_cleavage)


def non_specific_digest(seq, min_len, max_len):
    """Generate non specific digestion."""
    len_s = len(seq)
    for i in range(len_s + 1):
        for j in range(i + min_len, min(len_s + 1, i + max_len + 1)):
            if j <= len_s:
                yield seq[i:j]


def semi_specific_digest(seq, min_len, max_len, pre, not_post, miscleavages, methionine_cleavage):
    """Generate semi-specific digestion."""
    len_s, starts = len(seq), [0]
    methionine_cleavage = methionine_cleavage and seq[0] == "M"
    length_accepted = lambda x: x >= min_len and x <= max_len

    for i in range(len_s + 1):
        is_cleavage_site = seq[min([len_s - 1, i])] in pre and seq[min([len_s - 1, i + 1])] not in not_post
        ismethionine_cleavage_site = i == 0 and methionine_cleavage
        if i == len_s or is_cleavage_site or ismethionine_cleavage_site:
            # peptides with enzymatic C-terminal (both enzymatic and non-enzymatic N-terminal)
            start = starts[0]
            for j in range(start, min([i + 1, len_s])):
                len_p = min([i, len_s - 1]) - j + 1
                if length_accepted(len_p):
                    yield (seq[j : i + 1])
            starts.append(i + 1)
            methionine_cleaved = int(starts[0] == 0 and methionine_cleavage)
            if len(starts) > miscleavages + 1 + methionine_cleaved or i == len_s:
                starts = starts[1 + methionine_cleaved :]
        else:  # peptides with non enzymatic C-terminal
            for start in starts:
                len_p = i - start + 1
                if length_accepted(len_p) and i + 1 not in starts:
                    yield (seq[start : i + 1])


def full_digest(seq, min_len, max_len, pre, not_post, miscleavages, methionine_cleavage):
    """Generate full digestion."""
    len_s, starts = len(seq), [0]
    methionine_cleavage = methionine_cleavage and seq[0] == "M"
    length_accepted = lambda x: x >= min_len and x <= max_len

    cleavage_sites = [0] if methionine_cleavage else []
    cleavage_sites.extend([i for i in range(len_s) if seq[i] in pre and seq[min([len_s - 1, i + 1])] not in not_post])
    cleavage_sites.append(len_s)
    for i in cleavage_sites:
        for start in starts:
            len_p = i - start + 1
            if length_accepted(len_p):
                yield (seq[start : i + 1])
        starts.append(i + 1)
        methionine_cleaved = int(starts[0] == 0 and methionine_cleavage)
        if len(starts) > miscleavages + 1 + methionine_cleaved:
            starts = starts[1 + methionine_cleaved :]


def get_peptide_to_protein_map(
    fasta_file,
    db="concat",
    min_len=6,
    max_len=52,
    enzyme: str = "trypsin",
    digestion="full",
    miscleavages=2,
    methionine_cleavage=True,
    use_hash_key=False,
    special_aas=None,
    parse_id=parse_until_first_space,
):
    """Get peptide to protein map."""
    pre, not_post = cleavage_sites[enzyme]

    if pre is None:
        pre = ["K", "R"]
    if not_post is None:
        not_post = ["P"]
    if special_aas is None:
        special_aas = ["K", "R"]
    peptide_to_protein_map = collections.defaultdict(list)
    protein_to_seq_map = dict()
    for protein_idx, (protein, seq) in enumerate(read_fasta(fasta_file, db, parse_id, special_aas=special_aas)):
        if (protein_idx + 1) % 10000 == 0:
            logger.info(f"Digesting protein {protein_idx + 1}")
        seen_peptides = set()
        protein_to_seq_map[protein] = seq
        # for peptide in digestfast.get_digested_peptides(seq, min_len, max_len, pre, not_post, digestion,
        # miscleavages, methionine_cleavage):
        for peptide in get_digested_peptides(
            seq, min_len, max_len, pre, not_post, digestion, miscleavages, methionine_cleavage
        ):
            if use_hash_key:
                hash_key = peptide[:6]
            else:
                hash_key = peptide
            if hash_key not in seen_peptides:
                seen_peptides.add(hash_key)
                peptide_to_protein_map[hash_key].append(protein)

    if use_hash_key:
        return (peptide_to_protein_map, protein_to_seq_map)
    else:
        return peptide_to_protein_map


def get_peptide_to_protein_map_from_file(peptide_to_protein_map_file, use_hash_key=False):
    """Get peptide to protein map from file."""
    if use_hash_key:
        logger.warning("Hash key not supported yet, continuing without hash key...")
        use_hash_key = False
    peptide_to_protein_map = collections.defaultdict(list)
    reader = get_tsv_reader(peptide_to_protein_map_file)
    for i, row in enumerate(reader):
        if (i + 1) % 1000000 == 0:
            logger.info(f"Processing peptide  {i + 1}")

        peptide, proteins = row[0], row[1].split(";")
        if use_hash_key:
            sys.exit("Hash key not supported yet...")
            hash_key = peptide[:6]
        else:
            hash_key = peptide
        for protein in proteins:
            peptide_to_protein_map[hash_key].append(protein)
    return peptide_to_protein_map


def get_proteins(peptide_to_protein_map, peptide):
    """Get proteins."""
    peptide = peptide  # .replace("I", "L")
    if len(peptide_to_protein_map) == 2:
        hash_key = peptide[:6]
        proteins = list()
        if hash_key in peptide_to_protein_map[0]:
            for protein in peptide_to_protein_map[0][hash_key]:
                # TODO: This does not work correctly for full or partial digestion, since we might find the peptide
                # with the wrong number of enzymatic terminals
                if peptide in peptide_to_protein_map[1][protein]:
                    proteins.append(protein)
            proteins = sorted(proteins)
        return proteins
    else:
        return peptide_to_protein_map.get(peptide, [])


def get_all_proteins(peptide_to_protein_map):
    """Get all proteins."""
    seen_proteins = set()
    if len(peptide_to_protein_map) == 2:
        for _, proteins in peptide_to_protein_map[0].items():
            for protein in proteins:
                if protein not in seen_proteins:
                    seen_proteins.append(protein)
    else:
        for _, proteins in peptide_to_protein_map.items():
            for protein in proteins:
                if protein not in seen_proteins:
                    seen_proteins.append(protein)
    return list(seen_proteins)


def get_ibaq_peptide_to_protein_map(args):
    """Get ibaq peptide to protein map."""
    pre, not_post = cleavage_sites(args.enzyme)
    return get_peptide_to_protein_map(
        args.fasta,
        db="concat",
        digestion="full",
        min_len=max([6, args.min_length]),
        max_len=min([30, args.max_length]),
        enzyme=args.enzyme,
        miscleavages=0,
        methionine_cleavage=False,
        special_aas=list(args.special_aas),
    )


def get_num_ibaq_peptides_per_protein(args):
    """Get number of ibaq peptides per protein."""
    peptide_to_protein_map_ibaq = get_ibaq_peptide_to_protein_map(args)
    return get_num_peptides_per_protein(peptide_to_protein_map_ibaq)


def get_num_peptides_per_protein(peptide_to_protein_map):
    """Get number of peptides per protein."""
    num_peptides_per_protein = collections.defaultdict(int)
    for _, proteins in peptide_to_protein_map.items():
        for protein in proteins:
            num_peptides_per_protein[protein] += 1
    return num_peptides_per_protein


def is_enzymatic(aa1, aa2, pre=None, not_post=None, methionine_cleavage: bool = True) -> bool:
    """Check if is enzymatic."""
    if pre is None:
        pre = ["K", "R"]
    if not_post is None:
        not_post = ["P"]
    return aa1 == "-" or aa2 == "-" or (aa1 in pre and aa2 not in not_post) or (methionine_cleavage and aa1 == "M")


def has_miscleavage(seq, pre=None, not_post=None):
    """Check if peptide sequence has miscleavage."""
    if pre is None:
        pre = ["K", "R"]
    if not_post is None:
        not_post = ["P"]
    for i in range(len(seq) - 1):
        if is_enzymatic(seq[i], seq[i + 1], pre, not_post):
            return True
    return False


def get_tsv_reader(filename, delimiter="\t"):
    """Get tsv reader."""
    # Python 3
    if sys.version_info[0] >= 3:
        return csv.reader(open(filename, newline=""), delimiter=delimiter)
    # Python 2
    else:
        return csv.reader(open(filename, "rb"), delimiter=delimiter)


def get_tsv_writer(filename, delimiter="\t"):
    """Get tsv writer."""
    # Python 3
    if sys.version_info[0] >= 3:
        return csv.writer(open(filename, "w", newline=""), delimiter=delimiter)
    # Python 2
    else:
        return csv.writer(open(filename, "wb"), delimiter=delimiter)


if __name__ == "__main__":
    main(sys.argv[1:])
